"""This ui element represent the input, prompt and output of a callable step in the PromptMage."""

from nicegui import ui, run, app
from loguru import logger

from promptmage.mage import MageStep


RUNNING_ICON = "run_circle"
NOT_RUNNING_ICON = "circle"
SUCCESS_RUN_ICON = "check_circle"


def create_function_runner(step: MageStep):
    input_fields = {}
    system_prompt_field = None
    user_prompt_field = None
    result_field = None
    expansion_tab = (
        ui.expansion(f"Step: {step.name}", group="steps", icon=f"{NOT_RUNNING_ICON}")
        .classes("w-full")
        .style("width: 650px;")
    )
    # load prompt if available
    if step.prompt_name:
        prompt = step.get_prompt()
    else:
        prompt = None

    # run id given in app.storage, initialize with this data
    if app.storage.user.get("step_run_id"):
        step_run_id = app.storage.user.get("step_run_id")
        run_data = step.data_store.get_data(step_run_id)
        if run_data.step_name == step.name:
            prompt = run_data.prompt
            step.input_values = run_data.input_data
            step.result = run_data.output_data
            expansion_tab.props(f"icon={SUCCESS_RUN_ICON}")
            expansion_tab.update()
            del app.storage.user["step_run_id"]

    async def run_function():
        expansion_tab.props("icon=run_circle")
        expansion_tab.update()
        inputs = {name: field.value for name, field in input_fields.items()}
        if prompt is not None:
            prompt.system = system_prompt_field.value
            prompt.user = user_prompt_field.value
        result = await run.io_bound(step.execute, **inputs)
        expansion_tab.props(f"icon={SUCCESS_RUN_ICON}")
        expansion_tab.update()

    def set_prompt():
        nonlocal prompt
        system_prompt_field.update()
        user_prompt_field.update()
        prompt.system = system_prompt_field.value
        prompt.user = user_prompt_field.value
        prompt.version += 1
        step.set_prompt(prompt)

    def update_inputs():
        for name, field in input_fields.items():
            field.set_value(step.input_values[name])
            field.update()
        expansion_tab.props(f"icon={RUNNING_ICON}")
        expansion_tab.update()

    def update_results():
        result_field.set_content(f"{step.result}")
        result_field.update()

        expansion_tab.props(f"icon={SUCCESS_RUN_ICON}")
        expansion_tab.update()

    step.on_input_change(update_inputs)
    step.on_output_change(update_results)

    def build_ui():
        nonlocal user_prompt_field, system_prompt_field, result_field, expansion_tab
        with expansion_tab:
            with ui.card():
                ui.label(f"{step.name} - {step.step_id}")
                with ui.column().style(
                    "border: 1px solid #ddd; padding: 20px; border-radius: 10px; box-shadow: 0 4px 8px rgba(0, 0, 0, 0.1); margin-bottom: 20px; width: 1000px; height: 700px; overflow-y: auto;"
                ):
                    ui.label("Step Runner").style(
                        "font-weight: bold; font-size: 1.5em;"
                    )
                    with ui.row():
                        ui.label("Inputs:").style(
                            "margin-top: 20px; font-weight: bold;"
                        )
                        with ui.column().style("flex-grow: 1; margin-top: 20px;"):
                            for param in step.signature.parameters.values():
                                if param.name != "prompt":
                                    with ui.row():
                                        ui.label(f"{param.name}:").style(
                                            "width: 100px;"
                                        )
                                        input_fields[param.name] = ui.textarea(
                                            value=step.input_values[param.name]
                                        ).style("flex-grow: 1; overflow: auto;")
                        ui.label("Prompts:").style(
                            "margin-top: 20px; font-weight: bold;"
                        )
                        with ui.column().style("flex-grow: 1; margin-top: 20px;"):
                            with ui.row():
                                ui.label("System:").style("width: 100px;")
                                system_prompt_field = ui.textarea(
                                    value=(
                                        prompt.system
                                        if prompt
                                        else "No prompt supported"
                                    )
                                ).style("flex-grow: 1; overflow: auto;")
                            with ui.row():
                                ui.label("User:").style("width: 100px;")
                                user_prompt_field = ui.textarea(
                                    value=(
                                        prompt.user if prompt else "No prompt supported"
                                    )
                                ).style("flex-grow: 1; overflow: auto;")

                    with ui.row():
                        ui.button("Run", on_click=run_function).style(
                            "margin-top: 10px;"
                        )
                        ui.button("Save prompt", on_click=set_prompt).style(
                            "margin-top: 10px; margin-left: 10px;"
                        )
                    ui.separator()
                    ui.label("Result:").style("margin-top: 20px; font-weight: bold;")
                    result_field = ui.markdown(
                        f"{step.result}" if step.result else ""
                    ).style("margin-top: 20px; color: blue; height: 200px;")

    return build_ui



from promptmage import PromptMage
from nicegui import ui, app


def create_runs_view(mage: PromptMage):

    side_panel = ui.element("div").style(
        "position: fixed; top: 0; right: 0; width: 50%; height: 100%; background-color: #f0f0f0; transform: translateX(100%); transition: transform 0.3s ease; z-index: 1000; overflow-y: auto;"
    )

    def use_run_in_playground(step_run_id):
        app.storage.user["step_run_id"] = step_run_id
        ui.navigate.to("/")

    # Function to show the side panel with detailed information
    def show_side_panel(run_data):
        side_panel.clear()
        with side_panel:
            ui.button(">>", on_click=hide_side_panel).style(
                "margin: 20px; margin-bottom: 0px; margin-top: 100px;"
            )
            with ui.card().style(
                "padding: 20px; margin-right: 20px; margin-top: 20px; margin-bottom: 20px; margin-left: 20px"
            ):
                # display run data
                ui.label(f"Step Run ID: {run_data.step_run_id}")
                ui.label(f"Run ID: {run_data.run_id}")
                ui.label(f"Step Name: {run_data.step_name}")
                ui.label(f"Status: {run_data.status}")
                ui.label(f"Run Time: {run_data.run_time}")
                ui.label(f"Prompt: {run_data.prompt}")
                ui.label("Input Data:")
                for key, value in run_data.input_data.items():
                    ui.markdown(f"**{key}**: {value}")
                ui.label(f"Output Data: {run_data.output_data}")
                with ui.row():
                    ui.button(
                        "Use in playground",
                        on_click=lambda: use_run_in_playground(run_data.step_run_id),
                    )
        side_panel.style("transform:translateX(0%);")
        side_panel.update()

    # Function to hide the side panel
    def hide_side_panel():
        side_panel.clear()
        side_panel.style("transform:translateX(100%);")
        side_panel.update()

    def build_ui():
        ui.label("Runs").classes("text-2xl")
        runs = mage.get_run_data()

        # Main UI setup
        with ui.card().style("padding: 20px"):
            # Create a table with clickable rows
            columns = [
                {"name": "run_id", "label": "run_id", "field": "run_id"},
                {"name": "step_run_id", "label": "step_run_id", "field": "step_run_id"},
                {"name": "name", "label": "name", "field": "name", "sortable": True},
                {
                    "name": "status",
                    "label": "status",
                    "field": "status",
                    "sortable": True,
                },
                {
                    "name": "run_time",
                    "label": "run_time",
                    "field": "run_time",
                    "sortable": True,
                },
            ]

            rows = [
                {
                    "run_id": run_data.run_id,
                    "step_run_id": run_data.step_run_id,
                    "name": run_data.step_name,
                    "status": run_data.status,
                    "run_time": run_data.run_time,
                }
                for run_data in runs
            ]

            table = ui.table(
                columns=columns,
                rows=rows,
                pagination={
                    "rowsPerPage": 20,
                    "sortBy": "run_time",
                    "page": 1,
                    "descending": True,
                },
            )
            table.add_slot(
                "body-cell-status",
                """
                <q-td key="status" :props="props">
                    <q-badge :color="props.value === 'failed' ? 'red' : 'green'">
                        {{ props.value }}
                    </q-badge>
                </q-td>
                """,
            )

            def on_row_click(event):
                selected_run_index = event.args[-2]["step_run_id"]
                show_side_panel(
                    run_data=[r for r in runs if r.step_run_id == selected_run_index][
                        -1
                    ]
                )

            table.on("rowClick", on_row_click)

    return build_ui



"""Frontend module for PromptMage."""

from fastapi import FastAPI
from nicegui import ui

from promptmage import PromptMage
from promptmage.mage import MageStep

from .components import theme
from .components.main_runner import create_main_runner
from .components.step_runner import create_function_runner
from .components.runs_page import create_runs_view
from .components.prompts_page import create_prompts_view


class PromptMageFrontend:
    """A class that creates a frontend for a PromptMage instance."""

    def __init__(self, mage: PromptMage):
        self.mage = mage

    def init_from_api(self, fastapi_app: FastAPI) -> None:
        """Initialize the frontend from a FastAPI application."""

        @ui.page("/", title="PromptMage")
        def main_page():
            with theme.frame("Welcome to the PromptMage"):
                with ui.row().style("flex-wrap: wrap;"):
                    # Create a card for the mage
                    with ui.column().style("flex-wrap: wrap;"):
                        with ui.card():
                            ui.label("Run all steps")
                            create_main_runner(self.mage)()
                    # Create a card for each step
                    with ui.column().style("flex-wrap: wrap;"):
                        step: MageStep
                        for step in self.mage.steps.values():
                            create_function_runner(step)()

        @ui.page("/runs", title="PromptMage - Runs")
        def runs_page():
            with theme.frame("Runs Overview"):
                create_runs_view(self.mage)()

        @ui.page("/prompts", title="PromptMage - Prompts")
        def prompts_page():
            with theme.frame("Prompts Overview"):
                create_prompts_view(self.mage)()

        ui.run_with(
            fastapi_app,
            mount_path="/gui",  # NOTE this can be omitted if you want the paths passed to @ui.page to be at the root
            storage_secret="pick your private secret here",  # NOTE setting a secret is optional but allows for persistent storage per user
            dark=False,
            favicon="🧙",
        )



from nicegui import ui


def menu() -> None:
    with ui.button_group():
        ui.button("Playground", on_click=lambda: ui.navigate.to("/")).classes(
            replace="text-white"
        )
        ui.button("Runs", on_click=lambda: ui.navigate.to("/runs")).classes(
            replace="text-white"
        )
        ui.button("Prompts", on_click=lambda: ui.navigate.to("/prompts")).classes(
            replace="text-white"
        )



"""This module contains the DataStore class, which implements the storage and retrieval of data with different backends."""

from typing import Dict
from loguru import logger

from promptmage.storage import StorageBackend
from promptmage.run_data import RunData
from promptmage.exceptions import DataNotFoundException


class DataStore:
    """A class that stores and retrieves data with different backends."""

    def __init__(self, backend):
        self.backend: StorageBackend = backend

    def store_data(self, data: RunData):
        """Store data in the backend."""
        logger.info(f"Storing data: {data}")
        self.backend.store_data(data)

    def get_data(self, step_run_id: str) -> RunData:
        """Retrieve data from the backend."""
        logger.info(f"Retrieving data with ID: {step_run_id}")
        data = self.backend.get_data(step_run_id)
        if data:
            return data
        raise DataNotFoundException(step_run_id)

    def get_all_data(self) -> Dict:
        """Retrieve all data from the backend."""
        return self.backend.get_all_data()



"""This module contains the PromptStore class, which implements the storage and retrieval of prompts with different backends."""

from typing import List
from loguru import logger

from promptmage.storage import StorageBackend
from promptmage.prompt import Prompt
from promptmage.exceptions import PromptNotFoundException


class PromptStore:
    """A class that stores and retrieves prompts with different backends."""

    def __init__(self, backend):
        self.backend: StorageBackend = backend

    def store_prompt(self, prompt: Prompt):
        """Store a prompt in the backend."""
        logger.info(f"Storing prompt: {prompt}")
        self.backend.store_prompt(prompt)

    def get_prompt(self, prompt_id: str) -> Prompt:
        """Retrieve a prompt from the backend."""
        logger.info(f"Retrieving prompt with ID: {prompt_id}")
        try:
            return self.backend.get_prompt(prompt_id)
        except PromptNotFoundException:
            logger.error(
                f"Prompt with ID {prompt_id} not found, returning an empty prompt."
            )
            # return an empty prompt if the prompt is not found
            return Prompt(
                name=prompt_id,
                version=1,
                system="You are a helpful assistant.",
                user="",
                template_vars=[],
            )

    def get_prompts(self) -> List[Prompt]:
        """Retrieve all prompts from the backend."""
        return self.backend.get_prompts()

    def delete_prompt(self, prompt_id: str):
        """Delete a prompt from the backend."""
        logger.info(f"Deleting prompt with ID: {prompt_id}")
        self.backend.delete_prompt(prompt_id)



from .storage_backend import StorageBackend
from .sqlite_backend import SQLitePromptBackend, SQLiteDataBackend
from .file_backend import FileBackend
from .data_store import DataStore
from .memory_backend import InMemoryPromptBackend, InMemoryDataBackend
from .prompt_store import PromptStore

__all__ = [
    "StorageBackend",
    "SQLitePromptBackend",
    "SQLiteDataBackend",
    "FileBackend",
    "InMemoryPromptBackend",
    "InMemoryDataBackend",
    "DataStore",
    "PromptStore",
]



from nicegui import ui
from loguru import logger
from collections import defaultdict
from typing import List

from promptmage import PromptMage
from promptmage.prompt import Prompt


def create_prompts_view(mage: PromptMage):
    side_panel = ui.element("div").style(
        "position: fixed; top: 0; right: 0; width: 50%; height: 100%; background-color: #f0f0f0; transform: translateX(100%); transition: transform 0.3s ease; z-index: 1000; overflow-y: auto;"
    )

    # Function to show the side panel with detailed information
    def show_side_panel(prompts: List[Prompt]):
        # get prompt with highest version
        side_panel.clear()
        with side_panel:
            ui.button(">>", on_click=hide_side_panel).style(
                "margin: 20px; margin-bottom: 0px; margin-top: 100px;"
            )
            for prompt in sorted(prompts, key=lambda p: p.version, reverse=True):
                with ui.card().style(
                    "padding: 20px; margin-right: 20px; margin-top: 10px; margin-bottom: 10px; margin-left: 20px"
                ):
                    # display run data
                    ui.label(f"Prompt ID: {prompt.id}")
                    ui.label(f"Name: {prompt.name}")
                    ui.label(f"Version: {prompt.version}")
                    ui.label(f"System prompt: {prompt.system}")
                    ui.label(f"User prompt: {prompt.user}")
                    with ui.row():
                        ui.button(
                            "Use Prompt",
                            on_click=lambda prompt_id=prompt.id: use_prompt(prompt_id),
                        )
                        ui.button(
                            "Edit Prompt",
                            on_click=lambda prompt_id=prompt.id: edit_prompt(prompt_id),
                        )
                        ui.button(
                            "Delete Prompt",
                            on_click=lambda prompt_id=prompt.id: delete_prompt(
                                prompt_id
                            ),
                        )
        side_panel.style("transform:translateX(0%);")
        side_panel.update()

    def delete_prompt(prompt_id):
        mage.prompt_store.delete_prompt(prompt_id)

    def edit_prompt(prompt_id):
        logger.info(f"Editing prompt with ID: {prompt_id}. Not implemented yet.")

    def use_prompt(prompt_id):
        logger.info(f"Using prompt with ID: {prompt_id}. Not implemented yet.")

    # Function to hide the side panel
    def hide_side_panel():
        side_panel.clear()
        side_panel.style("transform:translateX(100%);")
        side_panel.update()

    def build_ui():
        ui.label("Prompts").classes("text-2xl")
        all_prompts = mage.prompt_store.get_prompts()
        # group them by name
        prompts = defaultdict(list)
        for prompt in all_prompts:
            if prompt.name in [step.prompt_name for step in mage.steps.values()]:
                prompts[prompt.name].append(prompt)

        # Main UI setup
        with ui.card().style("padding: 20px"):
            # Create a table with clickable rows
            columns = [
                {"name": "name", "label": "name", "field": "name", "sortable": True},
                {
                    "name": "versions",
                    "label": "Number of versions",
                    "field": "versions",
                    "sortable": True,
                },
            ]

            rows = [
                {
                    "name": name,
                    "versions": len(prompts_for_name),
                }
                for name, prompts_for_name in prompts.items()
            ]

            table = ui.table(columns=columns, rows=rows)

            def on_row_click(event):
                selected_run_index = event.args[-2]["name"]
                show_side_panel(prompts=prompts[selected_run_index])

            table.on("rowClick", on_row_click)

    return build_ui



"""This module contains the FileBackend class, which is a subclass of the StorageBackend class. It is used to store the data in a file on the local filesystem."""

from promptmage.storage.storage_backend import StorageBackend


class FileBackend(StorageBackend):
    """A class that stores the data in a file on the local filesystem."""

    def __init__(self, file_path: str):
        self.file_path = file_path



"""This module contains the SQLiteBackend class, which is a subclass of the StorageBackend class. It is used to store the data in a SQLite database."""

import sqlite3
import json
from loguru import logger
from pathlib import Path
from typing import List

from promptmage.prompt import Prompt
from promptmage.exceptions import PromptNotFoundException
from promptmage.run_data import RunData
from promptmage.storage.storage_backend import StorageBackend


class SQLitePromptBackend(StorageBackend):
    """A class that stores the prompts in a SQLite database.

    Attributes:
        db_path (str): The path to the SQLite database. Defaults to ".promptmage/promptmage.db".
    """

    def __init__(self, db_path: str | None = None):
        self.db_path = db_path if db_path else ".promptmage/promptmage.db"
        self._init_db()
        self._create_table()

    def _init_db(self):
        """Create the database if it doesn't exist."""
        conn = None
        try:
            conn = sqlite3.connect(self.db_path)
        except sqlite3.Error as e:
            logger.error(f"Error connecting to SQLite database: {e}")
        finally:
            if conn:
                conn.close()

    def _create_table(self):
        create_table_sql = """
        CREATE TABLE IF NOT EXISTS prompts (
            id TEXT PRIMARY KEY,
            name TEXT NOT NULL,
            system TEXT NOT NULL,
            user TEXT NOT NULL,
            version INTEGER NOT NULL,
            template_vars TEXT NOT NULL
        );
        """
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(create_table_sql)
            conn.commit()

    def store_prompt(self, prompt: Prompt):
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(
                "INSERT INTO prompts VALUES (?, ?, ?, ?, ?, ?)",
                (
                    prompt.id,
                    prompt.name,
                    prompt.system,
                    prompt.user,
                    prompt.version,
                    ",".join(prompt.template_vars),
                ),
            )
            conn.commit()

    def get_prompt(self, prompt_name: str) -> Prompt:
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute("SELECT * FROM prompts WHERE name=?", (prompt_name,))
            rows = cursor.fetchall()
            if len(rows) == 0:
                raise PromptNotFoundException(
                    f"Prompt with name {prompt_name} not found."
                )
            # select the latest version
            row = sorted(rows, key=lambda row: row[4], reverse=True)[0]
            return Prompt(
                id=row[0],
                name=row[1],
                system=row[2],
                user=row[3],
                version=row[4],
                template_vars=row[5].split(","),
            )

    def get_prompts(self) -> List[Prompt]:
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute("SELECT * FROM prompts")
            rows = cursor.fetchall()
            prompts = []
            for row in rows:
                prompts.append(
                    Prompt(
                        id=row[0],
                        name=row[1],
                        system=row[2],
                        user=row[3],
                        version=row[4],
                        template_vars=row[5].split(","),
                    )
                )
            return prompts

    def delete_prompt(self, prompt_id: str):
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute("DELETE FROM prompts WHERE id=?", (prompt_id,))
            conn.commit()


class SQLiteDataBackend(StorageBackend):
    """A class that stores the data in a SQLite database.

    Attributes:
        db_path (str): The path to the SQLite database. Defaults to ".promptmage/promptmage.db".
    """

    def __init__(self, db_path: str | None = None):
        self.db_path = db_path if db_path else ".promptmage/promptmage.db"
        self._init_db()
        self._create_table()

    def _init_db(self):
        conn = None
        try:
            conn = sqlite3.connect(self.db_path)
        except sqlite3.Error as e:
            logger.error(f"Error connecting to SQLite database: {e}")
        finally:
            if conn:
                conn.close()

    def _create_table(self):
        create_table_sql = """
        CREATE TABLE IF NOT EXISTS data (
            step_run_id TEXT PRIMARY KEY,
            run_time TEXT NOT NULL,
            step_name TEXT NOT NULL,
            run_id TEXT,
            status TEXT NOT NULL,
            prompt TEXT,
            input_data TEXT,
            output_data TEXT
        );
        """
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(create_table_sql)
            conn.commit()

    def store_data(self, run_data: RunData):
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(
                "INSERT INTO data VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
                (
                    run_data.step_run_id,
                    run_data.run_time,
                    run_data.step_name,
                    run_data.run_id,
                    run_data.status,
                    json.dumps(run_data.prompt.to_dict()) if run_data.prompt else None,
                    json.dumps(run_data.input_data),
                    json.dumps(run_data.output_data),
                ),
            )
            conn.commit()

    def get_data(self, step_run_id: str) -> RunData:
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute("SELECT * FROM data WHERE step_run_id=?", (step_run_id,))
            row = cursor.fetchone()
            if row is None:
                return None
            return RunData(
                step_run_id=row[0],
                run_time=row[1],
                step_name=row[2],
                run_id=row[3],
                status=row[4],
                prompt=Prompt.from_dict(json.loads(row[5])) if row[5] else None,
                input_data=json.loads(row[6]),
                output_data=json.loads(row[7]),
            )

    def get_all_data(self) -> List[RunData]:
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute("SELECT * FROM data")
            rows = cursor.fetchall()
            run_data = []
            for row in rows:
                r = RunData(
                    step_run_id=row[0],
                    run_time=row[1],
                    step_name=row[2],
                    run_id=row[3],
                    status=row[4],
                    prompt=Prompt.from_dict(json.loads(row[5])) if row[5] else None,
                    input_data=json.loads(row[6]),
                    output_data=json.loads(row[7]),
                )
                run_data.append(r)

            return run_data



"""This module contains the API for the PromptMage package."""

import inspect
import pkg_resources
from loguru import logger
from pathlib import Path
from pydantic import BaseModel

from fastapi import FastAPI, Path, Query
from fastapi.staticfiles import StaticFiles
from fastapi.responses import HTMLResponse
from fastapi.middleware.cors import CORSMiddleware


from promptmage import PromptMage


class PromptMageAPI:
    """A class that creates a FastAPI application to serve a PromptMage instance."""

    def __init__(self, mage: PromptMage):
        self.mage = mage

    def get_app(self) -> FastAPI:
        """Create a FastAPI application to serve the PromptMage instance."""
        app = FastAPI(
            title=f"PromptMage API: {self.mage.name}", description="API for PromptMage."
        )

        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

        # create index endpoint
        @app.get("/")
        async def index():
            return HTMLResponse(
                """<h1>Welcome to the PromptMage</h1>"""
                """<h2>To see the API docs 📖, go to <a href='/docs'>/docs</a></h2>"""
                """<h2>To access the API 🔮, go to <a href='/api'>/api</a></h2>"""
                """<h2>To access the GUI 🧙, go to <a href='/gui'>/gui</a></h2>"""
            )

        @app.get("/api")
        async def root():
            return HTMLResponse("<h1>Welcome to the PromptMage API</h1>")

        @app.get("/api/prompts")
        async def list_prompts():
            return self.mage.prompt_store.get_prompts()

        @app.get("/api/data")
        async def list_data():
            return self.mage.data_store.get_all_data()

        # create the endpoints for each step
        step_list = []
        for step_name, step in self.mage.steps.items():
            signature = inspect.signature(step.func)
            path = f"/api/{step_name}"
            params, path_variables = self._parameters_from_signature(signature)
            path += path_variables

            # Update the signature for the endpoint function
            new_signature = signature.replace(parameters=params)
            endpoint_func = self.create_endpoint_function(step.func, new_signature)
            setattr(
                endpoint_func, "__signature__", new_signature
            )  # Update the signature for FastAPI to recognize

            # Add the route to FastAPI
            app.add_api_route(
                path, endpoint_func, methods=["GET"], response_model=EndpointResponse
            )
            step_list.append({"name": step_name, "path": path})

        # create an endpoint to run the full dependency graph of the flow
        run_function = self.mage.get_run_function()
        signature = inspect.signature(run_function)
        path = "/api/run_flow"
        params, path_variables = self._parameters_from_signature(signature)
        path += path_variables
        new_signature = signature.replace(parameters=params)
        endpoint_func = self.create_endpoint_function(run_function, new_signature)
        setattr(endpoint_func, "__signature__", new_signature)
        app.add_api_route(
            path, endpoint_func, methods=["GET"], response_model=EndpointResponse
        )

        # add a route to list all available steps with their names and input variables
        @app.get("/api/steps")
        async def list_steps():
            return step_list

        static_files_path = pkg_resources.resource_filename("promptmage", "static/")
        app.mount(
            "/static/",
            StaticFiles(directory=static_files_path, html=True),
            name="static",
        )
        return app

    def _parameters_from_signature(self, signature):
        params = []
        path = ""
        for name, param in signature.parameters.items():
            # ignore prompt parameter
            if name == "prompt":
                continue
            if param.default is inspect.Parameter.empty:
                # Assume required parameters are path parameters
                new_param = inspect.Parameter(
                    name,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=Path(..., description=f"Path parameter `{name}`"),
                    annotation=param.annotation,
                )
                path += f"/{{{name}}}"  # Add to the path
            else:
                # Parameters with defaults are query parameters
                new_param = inspect.Parameter(
                    name,
                    kind=inspect.Parameter.KEYWORD_ONLY,
                    default=Query(
                        param.default, description=f"Query parameter `{name}`"
                    ),
                    annotation=param.annotation,
                )
            params.append(new_param)
        return params, path

    def create_endpoint_function(self, func, params):
        # Define the endpoint function using dynamic parameters
        async def endpoint(*args, **kwargs):
            try:
                result = func(*args, **kwargs)
                return EndpointResponse(
                    name=f"{func.__name__}",
                    status=200,
                    message="Success",
                    result=str(result),
                )
            except Exception as e:
                logger.info(f"Failed to call {func.__name__}")
                return EndpointResponse(
                    name=f"{func.__name__}",
                    status=500,
                    message=f"Error when calling {func.__name__}: {e}",
                )

        return endpoint


class EndpointResponse(BaseModel):
    name: str
    status: int = 500
    message: str = "Internal Server Error"
    result: str | None = None
